package com.ibm.tools.attach.javaSE;
/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2009, 2011  All Rights Reserved
 */

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import com.ibm.tools.attach.VirtualMachineDescriptor;
import com.ibm.tools.attach.spi.AttachProvider;

abstract class CommonDirectory {
	private static final String ATTACH_LOCK = "_attachlock"; //$NON-NLS-1$
	private static final String COM_IBM_TOOLS_ATTACH_DIRECTORY = "com.ibm.tools.attach.directory"; //$NON-NLS-1$
	private static final int COMMON_DIRECTORY_PERMISSIONS = 01777; /* allow anyone to create directories, but only owner can delete */
	private static final int COMMON_LOCK_FILE_PERMISSIONS = 0777; /* allow anyone to create and use the file */
	private static final String MASTER_LOCKFILE = "_master"; //$NON-NLS-1$
	private static final String MASTER_NOTIFIER = "_notifier"; //$NON-NLS-1$
	static final int SEMAPHORE_OKAY = 0;
	private static final String TRASH_PREFIX = ".trash_"; //$NON-NLS-1$

	static class syncObject {}
	private static final syncObject accessorMutex = new syncObject();
	private static FileLock attachLock;
	private static File commonDirFile; /* file where all IPC files are held */
	private static FileLock masterLock;
	private static String semaphoreId;
	/**
	 * default name of directories where VMs place their advertisements
	 */
	private static String systemTmpDir = IPC.getTmpDir();
	/**
	 * Indicates the control file is newer than the semaphore
	 */
	static final int J9PORT_INFO_SHSEM_OPENED_STALE = 103;
	/**
	 * commonDir holds the IPC directories, including those holding advertisements and semaphores
	 * @param dir File object for the IPC directory.
	 */
	private  static void setCommonDirFileObject(File dir) {
		synchronized (accessorMutex) {
			commonDirFile = dir;
		}
	}

	/**
	 * @return File object for the IPC directory.
	 */
	static  File getCommonDirFileObject() {
		synchronized (accessorMutex) { /* Jazz 27019: lazy initialisation of the object */
			if (null == commonDirFile) {
				initialize();
			}
			return commonDirFile;
		}
	}

	/**
	 * @return Path to the IPC directory.
	 * @note this strips trailing directory separators
	 */
	static  String getCommonDirPath() {
		return getCommonDirFileObject().getAbsolutePath();
	}

	private static void initialize() {
		/* Set the common directory used by all VMs.  Defaults to /tmp/.com_ibm_tools_attach */
		String ipcDirProperty = System.getProperty(COM_IBM_TOOLS_ATTACH_DIRECTORY,
				(new File(systemTmpDir,".com_ibm_tools_attach")).getPath()); //$NON-NLS-1$
		if (IPC.loggingEnabled ) {
			IPC.logMessage("IPC Directory", ipcDirProperty); //$NON-NLS-1$
		}
		File cd = new File(ipcDirProperty);
		setCommonDirFileObject(cd);
	}

	/**
	 * Sets up the common directory
	 * @throws IOException if common directory is not writable
	 */
	static void prepareCommonDirectory() throws IOException {
		File cd = getCommonDirFileObject();
		if (IPC.loggingEnabled ) {
			IPC.logMessage("createDirectoryAndSemaphore ", cd.getAbsolutePath()); //$NON-NLS-1$
		}
		if (cd.exists()) {
			if (!cd.canWrite()) {
				IPC.logMessage("could not write ", cd.getAbsolutePath()); //$NON-NLS-1$
				throw new IOException(cd.getAbsolutePath());
			} else if (!cd.isDirectory()) {
				if (!cd.delete()) { /* file where a directory should be */
					IPC.logMessage("could not delete ", cd.getAbsolutePath()); //$NON-NLS-1$
					throw new IOException(cd.getAbsolutePath());
				} else {
					IPC.logMessage("deleted ", cd.getAbsolutePath()); //$NON-NLS-1$
				}
				IPC.mkdirWithPermissions(cd.getAbsolutePath(), COMMON_DIRECTORY_PERMISSIONS);
			}
		} else {
			IPC.mkdirWithPermissions(cd.getAbsolutePath(), COMMON_DIRECTORY_PERMISSIONS);
		}
	}
	/**
	 * Lock the master lock file. Create the lockfile if necessary.
	 * @throws IOException if the file is already locked.
	 */
	static void obtainMasterLock() throws IOException {
		getMasterLock().lockFile(true);
	}

	/**
	 * non-blocking lock the master lockfile. Create the lockfile if necessary.
	 * @return true if lock obtained
	 * @throws IOException if file already locked
	 */
	static boolean tryObtainMasterLock() throws IOException {
		return getMasterLock().lockFile(false);
	}

	/**
	 * Release the lock on the master lock file
	 */
	static void releaseMasterLock() {
		getMasterLock().unlockFile();
	}

	/**
	 * Lock the attach lockfile. Create the lockfile if necessary.
	 * @throws IOException if file already locked
	 */

	static void obtainAttachLock() throws IOException {
		getAttachLock().lockFile(true);
	}

	/**
	 * Release the lock on the attach lock file
	 */
	static void releaseAttachLock() {
		try {
			getAttachLock().unlockFile();
		} catch (NullPointerException e) {
			return;
		}
	}

	/**
	 * Create the control file for the semaphore
	 * @throws IOException
	 */
	static void createNotificationFile() throws IOException {
		File notifierFile = new File(getCommonDirFileObject(), MASTER_NOTIFIER);
		if (notifierFile.createNewFile()) {
			IPC.chmod(notifierFile.getAbsolutePath(), COMMON_LOCK_FILE_PERMISSIONS);
		}
	}

	/**
	 * @param semId name of the semaphore, not including control directory name
	 */
	static void setSemaphoreName(String semId) {
		synchronized (accessorMutex) {
			semaphoreId = semId;
		}
	}

	/**
	 * @return name of the semaphore, not including control directory name
	 */
	static String getSemaphoreName() {
		synchronized (accessorMutex) {
			if (null == semaphoreId) {
				semaphoreId = MASTER_NOTIFIER;
			}
		}
		return semaphoreId;
	}

	/**
	 * @param obtainLock
	 * @return name of semaphore
	 * @throws IOException if mkdir fails
	 * Caller is responsible for ensuring that the master lockfile is held.
	 */
	static String openSemaphore() throws IOException {
		String semName = getSemaphoreName(); /*[PR Jazz 30075] replace inline assignment for clarity */
		int status = IPC.openSemaphore(getCommonDirFileObject().getAbsolutePath(), semName);
		// K0538 = semaphore {0} status= {1}
		if (SEMAPHORE_OKAY != status) {
			throw new IOException(com.ibm.oti.util.Msg.getString("K0538" , semName, Integer.valueOf(status)));  //$NON-NLS-1$
		}
		return semName;
	}

	/**
	 * close the semaphore and reopen it.
	 * @return 0 if success
	 */
	static int reopenSemaphore() {
		int status = 0;
		IPC.logMessage("reopenSemaphore"); //$NON-NLS-1$
		closeSemaphore();
		status = IPC.openSemaphore(getCommonDirFileObject().getAbsolutePath(), getSemaphoreName());
		return status;
	}

	/**
	 * wait for a post on the semaphore for this VM. Use notifyVm() to do the post
	 * @param myVmId for diagnostics
	 * @return 0 if success.
	 */
	static int waitSemaphore(String myVmId) {

		int status = IPC.waitSemaphore();
		if (SEMAPHORE_OKAY != status) {
			IPC.logMessage("waitSemaphore status != JNI_OK, ==", status); //$NON-NLS-1$
		}
		return status;
	}

	/**
	 * Open the semaphore, post to it, and close it
	 * @param numberOfTargets number of times to post to the semaphore
	 * @return 0 on success
	 */
	static int notifyVm(int numberOfTargets) {
		if (IPC.loggingEnabled ) {
			IPC.logMessage("notifyVm ", numberOfTargets, " targets"); //$NON-NLS-1$ //$NON-NLS-2$
		}
		return IPC.notifyVm(getCommonDirFileObject().getAbsolutePath(), getSemaphoreName(), numberOfTargets);
	}

	/**
	 * Open the semaphore, decrement it without blocking to it, and close it
	 * @param numberOfTargets number of times to decrement to the semaphore
	 * @return 0 on success
	 */
	static int cancelNotify(int numberOfTargets) {
		return IPC.cancelNotify(getCommonDirPath(), getSemaphoreName(), numberOfTargets);
	}

	/**
	 * close but do not destroy this VM's notification semaphore
	 */
	static void closeSemaphore() {

		IPC.closeSemaphore();
	}

	/**
	 * close and destroy the semaphore if possible
	 */
	static void destroySemaphore() {
		IPC.destroySemaphore();
	}

	/**
	 * @return Number of VMs which are advertising, including this VM
	 */
	static int getNumberOfTargets() {
		int NumberOfTargets =0;
		List<AttachProvider> providers = AttachProvider.providers();
		if ((null == providers) || (providers.size() == 0)) {
			return 0;
		}
		Iterator<AttachProvider> ip = providers.iterator();
		while (ip.hasNext()) {
			AttachProvider tempProvider = ip.next();
			if (tempProvider.getClass().equals(AttachProviderImpl.class)) {
				List<VirtualMachineDescriptor> vmds = tempProvider.listVirtualMachines();
				if (null != vmds) {
					NumberOfTargets = vmds.size();
				}
			}
		}
		return NumberOfTargets;
	}

	static int countTargetDirectories () {
		File dir= getCommonDirFileObject();

		if (!dir.isDirectory()) {
			return 0;
		}
		File[] vmDirs = dir.listFiles();
		if (null == vmDirs) {
			return 0;
		} else {
			int count = 0;
			for (int i=0; i < vmDirs.length; ++i) {
				String dirMemberName = vmDirs[i].getName();
				if (dirMemberName.startsWith(TRASH_PREFIX) || isCommonControlFile(dirMemberName)) {
					continue;
				}
				++count;
			}
			return count;
		}
	}

	static private boolean isCommonControlFile(String dirMemberName) {
		return (ATTACH_LOCK.equalsIgnoreCase(dirMemberName) || MASTER_LOCKFILE.equalsIgnoreCase(dirMemberName)
				|| MASTER_NOTIFIER.equalsIgnoreCase(dirMemberName));
	}

	/*
	 * look for leftover files and directories from previous VMs
	 */
	@SuppressWarnings("synthetic-access")
	static void deleteStaleDirectories(String myId) {
		long myUid = IPC.getUid();
		File[] vmDirs = getCommonDirFileObject().listFiles(new DirectorySampler());
		if (null == vmDirs) {
			return;
		}
		for (File dirMember: vmDirs) {
			if (AttachHandler.isAttachApiTerminated()) {
				break;
			}

			if (!isFileOwnedByUid(dirMember, myUid)) {
				continue;
			}
			String dirMemberName = dirMember.getName();
			if (IPC.loggingEnabled ) {
				IPC.logMessage("deleteStaleDirectories checking ", dirMemberName); //$NON-NLS-1$
			}
			if (dirMember.isFile()) {
				if (isCommonControlFile(dirMemberName)) {
					continue;
				}
				if (dirMemberName.startsWith(TRASH_PREFIX)) {
					IPC.logMessage("delete trash file ", dirMemberName); //$NON-NLS-1$
					if (!dirMember.delete()) {
						IPC.logMessage("error deleting ", dirMemberName); //$NON-NLS-1$
					}
				}
			} else { /* the member is a directory */
				if (dirMemberName.equalsIgnoreCase(myId)) {
					continue; /* Jazz 26865: no need to check myself */
				}
				long pid;
				if (Character.isDigit(dirMemberName.charAt(0))) {
					/*
					 * directory name is probably numeric.  Doing a quick test and catching the exception
					 * is faster than doing a full regex match.
					 */
					try {
						long uid;
						pid = Long.parseLong(dirMemberName);
						uid = CommonDirectory.getFileOwner(dirMember.getAbsolutePath());
						if (((0 != myUid) && (uid != myUid))) {
							/*
							 * If I am not running as root and the target is owned by another UID, then ignore the file.
							 */
							pid = 0;
						}
					} catch (NumberFormatException e) {
						pid = getPidFromFile(dirMember, myUid);
					}
				} else {
					pid = getPidFromFile(dirMember, myUid);
				}
				if ((0 != pid) /* the PID is valid and directory is owned by me or I am root */
						&& !IPC.processExists(pid)) {
					IPC.logMessage("delete defunct directory ", dirMemberName); //$NON-NLS-1$
					Advertisement.deleteAdvertisementDir(dirMemberName);
				}
			}
		}
	}

	/**
	 * Check if the file is owned by the UID.  Note that UID 0 "owns" all files.
	 * @param f File or directory
	 * @param myUid user UID.
	 * @return true if the uid owns the file or uid == 0.
	 */
	static boolean isFileOwnedByUid(File f, long myUid) {
		return (0 == myUid) || (myUid == CommonDirectory.getFileOwner(f.getAbsolutePath()));
	}

		/**
		 * reads the process ID from the advertisement file.
		 * Returns 0 (illegal PID) if the advertisement file does not exist
		 * or the process is not owned by the current user.
		 * @param dirMember directory which should hold an advertisement file
		 * @return the target process's PID or 0
		 */
	private static long getPidFromFile(File dirMember, long myUid) {
		long pid = 0;
		File advertFile = new File(dirMember, Advertisement.getFilename());
		if (!advertFile.exists() || (0 == advertFile.length())) {
			IPC.logMessage("delete stale directory ", dirMember.getName()); //$NON-NLS-1$
			Advertisement.deleteAdvertisementDir(dirMember.getName());
		}
		Advertisement advert;
		try {
			FileInputStream propStream = new FileInputStream(advertFile);
			advert = Advertisement.readAdvertisementFile(propStream); /* throws IOException if file cannot be read */
			propStream.close();
		} catch (IOException e) {
			return 0;
		}
		pid = advert.getProcessId();
		long uid = advert.getUid();
		if (IPC.loggingEnabled) {
			IPC.logMessage("getPidFromFile pid = ", (int) pid, dirMember.getName()); //$NON-NLS-1$
		}
		/*advertisement is from an older version or is corrupt, or claims to be owned by root.  Get the owner via file stat ]*/
		if (0 == uid) {
			uid = CommonDirectory.getFileOwner(dirMember.getAbsolutePath());
		}
		if (IPC.loggingEnabled) {
			IPC.logMessage("getPidFromFile uid = ", (int) uid); //$NON-NLS-1$
		}
		if (((0 != myUid) && (uid != myUid))) {
			/*
			 * If I am not running as root and the target is owned by another UID, then ignore the file.
			 */
			pid = 0;
		}
		return pid;
	}

	private static FileLock getAttachLock() {
		synchronized (accessorMutex) {
			if (null == attachLock) { /* hang on to a copy since we need it at shutdown */
				attachLock = new FileLock((new File(getCommonDirFileObject(), ATTACH_LOCK)).getAbsolutePath(), COMMON_LOCK_FILE_PERMISSIONS);
			}
		}
		return attachLock;
	}

	private static FileLock getMasterLock() {
		synchronized (accessorMutex) {
			if (null == masterLock) {
				masterLock = new FileLock((new File(getCommonDirFileObject(), MASTER_LOCKFILE)).getAbsolutePath(), COMMON_LOCK_FILE_PERMISSIONS);
			}
		}
		return masterLock;
	}

	static native long getFileOwner(String path);

	private static class DirectorySampler implements FileFilter {

		private int acceptCount = 16;
		private long skip;
		private long range = 2;
		public boolean accept(File candidate) {
			if (acceptCount > 0) { /* accept the first N files unconditionally */
				--acceptCount;
				return true;
			}
			/* then accept one, reject a random number (possibly 0), accept one, reject a larger random number, accept one, and so on. */
			if (skip > 0) {
				--skip;
				return false;
			} else {
				skip = System.currentTimeMillis() % range;
				range *= 2;
				return true;
			}
		}

	}
}
